{ pkgs, ... }: {
  virtualisation.vmVariant = {
    # let us log in
    users.mutableUsers = false;
    users.users.root.hashedPassword = "";
    services.openssh = {
      enable = true;
      settings = {
        PermitRootLogin = "yes";
        PermitEmptyPasswords = "yes";
        UsePAM = "no";
      };
    };

    # automatically log in
    services.getty.autologinUser = "root";

    # access to convenient things
    environment.systemPackages = with pkgs; [ w3m python3 ];
    nix.extraOptions = ''
      extra-experimental-features = nix-command flakes
    '';

    # no graphics. see nixos-shell
    virtualisation = {
      graphics = false;
      qemu.consoles = [ "tty0" "hvc0" ];
      qemu.options = [
        "-serial null"
        "-device virtio-serial"
        "-chardev stdio,mux=on,id=char0,signal=off"
        "-mon chardev=char0,mode=readline"
        "-device virtconsole,chardev=char0,nr=0"
      ];
      forwardPorts = [
        {
          from = "host";
          host.port = 2222;
          guest.port = 22;
        }
      ];
    };
  };
}
